# The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime
from trytond.model import fields
from trytond.pool import PoolMeta, Pool
from trytond.pyson import Eval

__all__ = ['ShipmentOut', 'ShipmentInternal']


class ShipmentOut:
    __name__ = 'stock.shipment.out'
    __metaclass__ = PoolMeta

    unit_loads = fields.Function(
        fields.One2Many('stock.unit_load', None, 'Unit loads',
                        states={'readonly': Eval('state') != 'draft'},
                        depends=['state']),
        'get_unit_loads', setter='set_unit_loads', searcher='search_unit_loads')

    @classmethod
    def __setup__(cls):
        super(ShipmentOut, cls).__setup__()
        for _field_name in ('inventory_moves', 'outgoing_moves'):
            _field = getattr(cls, _field_name)
            if _field.states.get('readonly'):
                _field.states['readonly'] |= Eval('unit_loads')
            else:
                _field.states['readonly'] = Eval('unit_loads')
            if 'unit_loads' not in _field.depends:
                _field.depends.append('unit_loads')

    def get_unit_loads(self, name=None):
        if not self.outgoing_moves:
            return []
        uls = set(m.unit_load.id for m in self.outgoing_moves if m.unit_load)
        return list(uls)

    @classmethod
    def set_unit_loads(cls, records, name, value):
        pass

    @classmethod
    def search_unit_loads(cls, name, clause):
        _field = 'moves.unit_load'
        if '.' in name:
            _field += '.%s' % name[10:]
        return [(_field, ) + tuple(clause[1:])]

    @fields.depends('unit_loads', 'outgoing_moves', 'warehouse_output',
                    'customer_location', 'end_date')
    def on_change_unit_loads(self):
        moves = []
        ul_ids = [ul.id for ul in self.unit_loads]
        if self.outgoing_moves:
            moves.extend([m for m in self.outgoing_moves if m.unit_load.id in ul_ids])
        for ul in self.unit_loads:
            if ul.id in [m.unit_load.id for m in moves]:
                continue
            new_moves = ul._move(self.customer_location,
                                 self.end_date,
                                 from_location=self.warehouse_output)
            moves.extend(new_moves)
        self.outgoing_moves = moves

    def _get_inventory_move(self, move):
        res = super(ShipmentOut, self)._get_inventory_move(move)
        if move.unit_load:
            location = move.unit_load.get_location([move.unit_load], type='storage')[move.unit_load.id]
            res.from_location = location
            res.unit_load = move.unit_load
        return res


class ShipmentInternal:
    __name__ = 'stock.shipment.internal'
    __metaclass__ = PoolMeta

    ul_quantity = fields.Function(
        fields.Float('ULs', digits=(16, 0)), 'get_ul_quantity')
    unit_loads = fields.Function(
        fields.One2Many('stock.unit_load', None, 'Unit loads',
                        states={'readonly': Eval('state') != 'draft'},
                        depends=['state']),
        'get_unit_loads', setter='set_unit_loads', searcher='search_unit_loads')

    @classmethod
    def __setup__(cls):
        super(ShipmentInternal, cls).__setup__()
        for _field_name in ('moves', ):
            _field = getattr(cls, _field_name)
            if _field.states.get('readonly'):
                _field.states['readonly'] |= Eval('unit_loads')
            else:
                _field.states['readonly'] = Eval('unit_loads')
            if 'unit_loads' not in _field.depends:
                _field.depends.append('unit_loads')

    def get_ul_quantity(self, name=None):
        if not self.unit_loads:
            return None
        return len(self.unit_loads)

    def get_unit_loads(self, name=None):
        if not self.moves:
            return []
        uls = set(m.unit_load.id for m in self.moves if m.unit_load)
        return list(uls)

    @classmethod
    def set_unit_loads(cls, records, name, value):
        pass

    @classmethod
    def search_unit_loads(cls, name, clause):
        _field = 'moves.unit_load'
        if '.' in name:
            _field += '.%s' % name[10:]
        return [(_field,) + tuple(clause[1:])]

    @fields.depends('unit_loads', 'moves', 'date_time_',
                    'from_location', 'to_location')
    def on_change_unit_loads(self):
        moves = []
        ul_ids = [ul.id for ul in self.unit_loads]
        if self.moves:
            moves.extend(
                [m for m in self.moves if m.unit_load.id in ul_ids])
        for ul in self.unit_loads:
            if ul.id in [m.unit_load.id for m in moves]:
                continue
            new_moves = ul._move(self.to_location,
                self.date_time_,
                from_location=self.from_location)
            moves.extend(new_moves)
        self.moves = moves
