# The COPYRIGHT file at the top level of this repository contains the full
# copyright notices and license terms.
import datetime
from trytond.pool import PoolMeta, Pool
from trytond.model import fields, ModelView
from trytond.pyson import PYSONEncoder, Date
from trytond.transaction import Transaction
from trytond.wizard import Wizard, StateView, Button, StateAction

__all__ = ['Move' 'UnitLoadsByLocationsStart',
           'UnitLoadsByLocations']


class Move:
    __name__ = 'stock.move'
    __metaclass__ = PoolMeta

    unit_load = fields.Many2One('stock.unit_load', 'Unit load',
                                ondelete='RESTRICT', select=True,
                                readonly=True)

    @classmethod
    def validate(cls, records):
        super(Move, cls).validate(records)
        for record in records:
            if record.unit_load and record.time_ is None:
                cls.raise_user_error('ul_time_required')

    @classmethod
    def __setup__(cls):
        super(Move, cls).__setup__()
        cls._error_messages.update({
            'ul_time_required': 'Time in Unit load moves is required.'})

    @classmethod
    def assign(cls, records):
        pool = Pool()
        Unitload = pool.get('stock.unit_load')

        super(Move, cls).assign(records)
        uls = [m.unit_load for m in records if m.unit_load]
        if uls:
            Unitload._set_internal_quantity(list(set(uls)))

    @classmethod
    def do(cls, records):
        pool = Pool()
        Unitload = pool.get('stock.unit_load')

        uls = [m.unit_load for m in records if m.state == 'draft' and m.unit_load]

        super(Move, cls).do(records)

        if uls:
            Unitload._set_internal_quantity(list(set(uls)))


class UnitLoadsByLocationsStart(ModelView):
    """Unit loads by Locations"""
    __name__ = 'stock.unit_loads_by_locations.start'

    forecast_date = fields.Date(
        'At Date', help=('Allow to compute expected '
            'stock quantities for this date.\n'
            '* An empty value is an infinite date in the future.\n'
            '* A date in the past will provide historical values.'))

    @staticmethod
    def default_forecast_date():
        Date_ = Pool().get('ir.date')
        return Date_.today()


class UnitLoadsByLocations(Wizard):
    """Unit loads by Locations"""
    __name__ = 'stock.unit_loads_by_locations'

    start = StateView('stock.unit_loads_by_locations.start',
                      'stock_unit_load.uls_by_locations_start_view_form',
                      [Button('Cancel', 'end', 'tryton-cancel'),
                       Button('Open', 'open', 'tryton-ok', True)])
    open = StateAction('stock_unit_load.act_uls_by_locations')

    def do_open(self, action):
        pool = Pool()
        Location = pool.get('stock.location')
        Lang = pool.get('ir.lang')

        context = {}
        context['locations'] = Transaction().context.get('active_ids')
        date = self.start.forecast_date or datetime.date.max
        context['stock_date_end'] = Date(date.year, date.month, date.day)
        action['pyson_context'] = PYSONEncoder().encode(context)

        locations = Location.browse(context['locations'])

        for code in [Transaction().language, 'en_US']:
            langs = Lang.search([
                    ('code', '=', code),
                    ])
            if langs:
                break
        lang = langs[0]
        date = Lang.strftime(date, lang.code, lang.date)

        action['name'] += ' - (%s) @ %s' % (
            ','.join(l.name for l in locations), date)
        return action, {}
